"""Test suite module for prance."""

__author__ = "Jens Finkhaeuser"
__copyright__ = "Copyright (c) 2016-2018 Jens Finkhaeuser"
__license__ = "MIT"
__all__ = ()


def _find_imports(*args):
    """
    Helper sorting the named modules into existing and not existing.
    """
    import importlib

    exists = {
        True: [],
        False: [],
    }

    for name in args:
        name = name.replace("-", "_")
        try:
            importlib.import_module(name)
            exists[True].append(name)
        except ImportError:
            exists[False].append(name)

    return exists


def none_of(*args):
    """
    Return true if none of the named modules exist, false otherwise.
    """
    exists = _find_imports(*args)
    return len(exists[True]) == 0


def _platform(platform, *args):
    """
    Helper for platform()
    """
    ret = False
    for arg in args:
        if arg[0] == "!":
            arg = arg[1:]
            if platform != arg:
                ret = True
                break
        elif platform == arg:
            ret = True
            break

    return ret


def platform(*args):
    """
    Return true if the current platform is in one of the given ones.

    Platforms may be specified with leading '!', in which case they
    a match returns False.
    """
    import sys

    return _platform(sys.platform, *args)



"""Test suite for prance.util.resolver ."""

__author__ = "Jens Finkhaeuser"
__copyright__ = "Copyright (c) 2016-2021 Jens Finkhaeuser"
__license__ = "MIT"
__all__ = ()


import pytest
from unittest.mock import patch

from prance.util import fs
from prance.util import resolver
from prance.util.url import ResolutionError

from . import none_of


def get_specs(fname):
    specs = fs.read_file(fname)

    from prance.util import formats

    specs = formats.parse_spec(specs, fname)

    return specs


def mock_get_petstore(*args, **kwargs):
    from .mock_response import MockResponse, PETSTORE_YAML

    return MockResponse(text=PETSTORE_YAML)


def recursion_limit_handler_none(limit, refstring, recursions):
    return None


@pytest.fixture
def externals_file():
    return get_specs("tests/specs/with_externals.yaml")


@pytest.fixture
def recursive_objs_file():
    return get_specs("tests/specs/recursive_objs.yaml")


@pytest.fixture
def recursive_files_file():
    return get_specs("tests/specs/recursive_files.yaml")


@pytest.fixture
def recursion_limit_file():
    return get_specs("tests/specs/recursion_limit.yaml")


@pytest.fixture
def recursion_limit_files_file():
    return get_specs("tests/specs/recursion_limit_files.yaml")


@pytest.fixture
def missing_file():
    return get_specs("tests/specs/missing_reference.yaml")


def test_resolver_noname(externals_file):
    res = resolver.RefResolver(externals_file)
    # Can't build a fragment URL without reference
    with pytest.raises(ResolutionError):
        res.resolve_references()


@patch("requests.get")
def test_resolver_named(mock_get, externals_file):
    mock_get.side_effect = mock_get_petstore

    import os.path
    from prance.util import fs

    res = resolver.RefResolver(
        externals_file, fs.abspath("tests/specs/with_externals.yaml")
    )
    res.resolve_references()


@patch("requests.get")
def test_resolver_missing_reference(mock_get, missing_file):
    mock_get.side_effect = mock_get_petstore

    import os.path

    res = resolver.RefResolver(
        missing_file, fs.abspath("tests/specs/missing_reference.yaml")
    )
    with pytest.raises(ResolutionError) as exc:
        res.resolve_references()

    assert str(exc.value).startswith("Cannot resolve")


@patch("requests.get")
def test_resolver_recursive_objects(mock_get, recursive_objs_file):
    mock_get.side_effect = mock_get_petstore

    # Recursive references to objects are a problem
    import os.path

    res = resolver.RefResolver(
        recursive_objs_file, fs.abspath("tests/specs/recursive_objs.yaml")
    )
    with pytest.raises(ResolutionError) as exc:
        res.resolve_references()

    assert str(exc.value).startswith("Recursion reached limit")


@patch("requests.get")
def test_resolver_recursive_files(mock_get, recursive_files_file):
    mock_get.side_effect = mock_get_petstore

    # Recursive references to files are not a problem
    import os.path

    res = resolver.RefResolver(
        recursive_files_file, fs.abspath("tests/specs/recursive_files.yaml")
    )
    res.resolve_references()


def test_recursion_limit_do_not_recurse_raise(recursion_limit_file):
    # Expect the default behaviour to raise.
    import os.path

    res = resolver.RefResolver(
        recursion_limit_file, fs.abspath("tests/specs/recursion_limit.yaml")
    )
    with pytest.raises(ResolutionError) as exc:
        res.resolve_references()

    assert str(exc.value).startswith("Recursion reached limit of 1")


def test_recursion_limit_do_not_recurse_ignore(recursion_limit_file):
    # If we overload the handler, we should not get an error but should
    # also simply not have the 'next' field - or it should be None
    import os.path

    res = resolver.RefResolver(
        recursion_limit_file,
        fs.abspath("tests/specs/recursion_limit.yaml"),
        recursion_limit_handler=recursion_limit_handler_none,
    )
    res.resolve_references()

    from prance.util import formats

    contents = formats.serialize_spec(res.specs, "foo.yaml")

    # The effect of returning None on recursion limit should be that
    # despite having recursion, the outermost reference to
    # definitions/Pet should get resolved.
    assert (
        "properties" in res.specs["paths"]["/pets"]["get"]["responses"]["200"]["schema"]
    )

    # However, the 'next' field should not be resolved.
    assert (
        res.specs["paths"]["/pets"]["get"]["responses"]["200"]["schema"]["properties"][
            "next"
        ]["schema"]
        is None
    )


def test_recursion_limit_set_limit_ignore(recursion_limit_file):
    # If we overload the handler, and set the recursion limit higher,
    # we should get nested Pet objects a few levels deep.

    import os.path

    res = resolver.RefResolver(
        recursion_limit_file,
        fs.abspath("tests/specs/recursion_limit.yaml"),
        recursion_limit=2,
        recursion_limit_handler=recursion_limit_handler_none,
    )
    res.resolve_references()

    from prance.util import formats

    contents = formats.serialize_spec(res.specs, "foo.yaml")

    # The effect of returning None on recursion limit should be that
    # despite having recursion, the outermost reference to
    # definitions/Pet should get resolved.
    assert (
        "properties" in res.specs["paths"]["/pets"]["get"]["responses"]["200"]["schema"]
    )

    # However, the 'next' field should be resolved due to the higher recursion limit
    next_field = res.specs["paths"]["/pets"]["get"]["responses"]["200"]["schema"][
        "properties"
    ]["next"]["schema"]
    assert next_field is not None

    # But the 'next' field of the 'next' field should not be resolved.
    assert next_field["properties"]["next"]["schema"] is None


def test_recursion_limit_do_not_recurse_raise_files(recursion_limit_files_file):
    # Expect the default behaviour to raise.
    import os.path

    res = resolver.RefResolver(
        recursion_limit_files_file, fs.abspath("tests/specs/recursion_limit_files.yaml")
    )
    with pytest.raises(ResolutionError) as exc:
        res.resolve_references()

    assert str(exc.value).startswith("Recursion reached limit of 1")


def test_recursion_limit_do_not_recurse_ignore_files(recursion_limit_files_file):
    # If we overload the handler, we should not get an error but should
    # also simply not have the 'next' field - or it should be None
    import os.path

    res = resolver.RefResolver(
        recursion_limit_files_file,
        fs.abspath("tests/specs/recursion_limit_files.yaml"),
        recursion_limit_handler=recursion_limit_handler_none,
    )
    res.resolve_references()

    from prance.util import formats

    contents = formats.serialize_spec(res.specs, "foo.yaml")

    # The effect of returning None on recursion limit should be that
    # despite having recursion, the outermost reference to
    # definitions/Pet should get resolved.
    assert (
        "properties" in res.specs["paths"]["/pets"]["get"]["responses"]["200"]["schema"]
    )

    # However, the 'next' field should not be resolved.
    assert (
        res.specs["paths"]["/pets"]["get"]["responses"]["200"]["schema"]["properties"][
            "next"
        ]["schema"]
        is None
    )


def test_recursion_limit_set_limit_ignore_files(recursion_limit_files_file):
    # If we overload the handler, and set the recursion limit higher,
    # we should get nested Pet objects a few levels deep.

    import os.path

    res = resolver.RefResolver(
        recursion_limit_files_file,
        fs.abspath("tests/specs/recursion_limit_files.yaml"),
        recursion_limit=2,
        recursion_limit_handler=recursion_limit_handler_none,
    )
    res.resolve_references()

    from prance.util import formats

    contents = formats.serialize_spec(res.specs, "foo.yaml")

    # The effect of returning None on recursion limit should be that
    # despite having recursion, the outermost reference to
    # definitions/Pet should get resolved.
    assert (
        "properties" in res.specs["paths"]["/pets"]["get"]["responses"]["200"]["schema"]
    )

    # However, the 'next' field should be resolved due to the higher recursion limit
    next_field = res.specs["paths"]["/pets"]["get"]["responses"]["200"]["schema"][
        "properties"
    ]["next"]["schema"]
    assert next_field is not None

    # But the 'next' field of the 'next' field should not be resolved.
    assert next_field["properties"]["next"]["schema"] is None


@patch("requests.get")
def test_issue_22_empty_path(mock_get, externals_file):
    mock_get.side_effect = mock_get_petstore

    # The raw externals file must have unresolved data
    assert len(externals_file["paths"]["/pets/{petId}"]["get"]["parameters"]) == 1
    param = externals_file["paths"]["/pets/{petId}"]["get"]["parameters"][0]

    assert "overwritten" in param
    assert "$ref" in param

    import os.path
    from prance.util import fs

    res = resolver.RefResolver(
        externals_file, fs.abspath("tests/specs/with_externals.yaml")
    )
    res.resolve_references()

    # The tests should resolve the reference, but the reference overwrites
    # all else.
    assert len(res.specs["paths"]["/pets/{petId}"]["get"]["parameters"]) == 1
    param = res.specs["paths"]["/pets/{petId}"]["get"]["parameters"][0]

    # Dereferenced keys must exist here
    assert "type" in param
    assert "description" in param
    assert "required" in param
    assert "in" in param
    assert "name" in param

    # Previously defined keys must not
    assert "overwritten" not in param
    assert "$ref" not in param


def test_issue_38_tilde_one():
    specs = get_specs("tests/specs/issue_38_a.yaml")
    res = resolver.RefResolver(specs, fs.abspath("tests/specs/issue_38_a.yaml"))
    res.resolve_references()

    path = res.specs["paths"]["/api/v2/vms"]
    assert "get" in path
    assert "operationId" in path["get"]
    assert "description" in path["get"]


@patch("requests.get")
def test_issue_23_partial_resolution_all(mock_get):
    mock_get.side_effect = mock_get_petstore

    specs = get_specs("tests/specs/with_externals.yaml")
    res = resolver.RefResolver(specs, fs.abspath("tests/specs/with_externals.yaml"))
    res.resolve_references()

    # By default, all externals need to be resolved.
    from prance.util.path import path_get

    val = path_get(res.specs, ("paths", "/pets", "get", "responses", "200", "schema"))
    assert "$ref" not in val

    val = path_get(
        res.specs, ("paths", "/pets", "get", "responses", "default", "schema")
    )
    assert "$ref" not in val

    val = path_get(
        res.specs, ("paths", "/pets", "post", "responses", "default", "schema")
    )
    assert "$ref" not in val

    val = path_get(res.specs, ("paths", "/pets/{petId}", "get", "parameters", 0))
    assert "$ref" not in val

    val = path_get(
        res.specs, ("paths", "/pets/{petId}", "get", "responses", "200", "schema")
    )
    assert "$ref" not in val

    val = path_get(
        res.specs, ("paths", "/pets/{petId}", "get", "responses", "203", "schema")
    )
    assert "$ref" not in val

    val = path_get(
        res.specs, ("paths", "/pets/{petId}", "get", "responses", "default", "schema")
    )
    assert "$ref" not in val


def test_issue_23_partial_resolution_internal():
    specs = get_specs("tests/specs/with_externals.yaml")
    res = resolver.RefResolver(
        specs,
        fs.abspath("tests/specs/with_externals.yaml"),
        resolve_types=resolver.RESOLVE_INTERNAL,
    )
    res.resolve_references()

    # By default, all externals need to be resolved.
    from prance.util.path import path_get

    val = path_get(res.specs, ("paths", "/pets", "get", "responses", "200", "schema"))
    assert "$ref" in val

    val = path_get(
        res.specs, ("paths", "/pets", "get", "responses", "default", "schema")
    )
    assert "$ref" in val

    val = path_get(
        res.specs, ("paths", "/pets", "post", "responses", "default", "schema")
    )
    assert "$ref" in val

    val = path_get(res.specs, ("paths", "/pets/{petId}", "get", "parameters", 0))
    assert "$ref" in val

    val = path_get(
        res.specs, ("paths", "/pets/{petId}", "get", "responses", "200", "schema")
    )
    assert "$ref" in val

    val = path_get(
        res.specs, ("paths", "/pets/{petId}", "get", "responses", "203", "schema")
    )
    assert "$ref" in val

    val = path_get(
        res.specs, ("paths", "/pets/{petId}", "get", "responses", "default", "schema")
    )
    assert "$ref" not in val


def test_issue_23_partial_resolution_files():
    specs = get_specs("tests/specs/with_externals.yaml")
    res = resolver.RefResolver(
        specs,
        fs.abspath("tests/specs/with_externals.yaml"),
        resolve_types=resolver.RESOLVE_FILES,
    )
    res.resolve_references()

    # By default, all externals need to be resolved.
    from prance.util.path import path_get

    val = path_get(res.specs, ("paths", "/pets", "get", "responses", "200", "schema"))
    assert "$ref" not in val

    val = path_get(
        res.specs, ("paths", "/pets", "get", "responses", "default", "schema")
    )
    assert "$ref" not in val

    val = path_get(
        res.specs, ("paths", "/pets", "post", "responses", "default", "schema")
    )
    assert "$ref" not in val

    val = path_get(res.specs, ("paths", "/pets/{petId}", "get", "parameters", 0))
    assert "$ref" not in val

    val = path_get(
        res.specs, ("paths", "/pets/{petId}", "get", "responses", "200", "schema")
    )
    assert "$ref" in val

    val = path_get(
        res.specs, ("paths", "/pets/{petId}", "get", "responses", "203", "schema")
    )
    assert "$ref" not in val

    val = path_get(
        res.specs, ("paths", "/pets/{petId}", "get", "responses", "default", "schema")
    )
    assert "$ref" in val


@pytest.mark.skipif(none_of("openapi-spec-validator"), reason="Missing backends")
def test_issue_65_partial_resolution_files():
    specs = """openapi: "3.0.0"
info:
  title: ''
  version: '1.0.0'
paths: {}
components:
    schemas:
        SampleArray:
            type: array
            items:
              $ref: '#/components/schemas/ItemType'

        ItemType:
          type: integer
"""
    from prance.util import formats

    specs = formats.parse_spec(specs, "issue_65.yaml")

    res = resolver.RefResolver(
        specs, fs.abspath("issue_65.yaml"), resolve_types=resolver.RESOLVE_FILES
    )
    res.resolve_references()

    from prance.util.path import path_get

    val = path_get(res.specs, ("components", "schemas", "SampleArray", "items"))
    assert "$ref" in val


@patch("requests.get")
def test_issue_23_partial_resolution_http(mock_get):
    mock_get.side_effect = mock_get_petstore

    specs = get_specs("tests/specs/with_externals.yaml")
    res = resolver.RefResolver(
        specs,
        fs.abspath("tests/specs/with_externals.yaml"),
        resolve_types=resolver.RESOLVE_HTTP,
    )
    res.resolve_references()

    # By default, all externals need to be resolved.
    from prance.util.path import path_get

    val = path_get(res.specs, ("paths", "/pets", "get", "responses", "200", "schema"))
    assert "$ref" in val

    val = path_get(
        res.specs, ("paths", "/pets", "get", "responses", "default", "schema")
    )
    assert "$ref" in val

    val = path_get(
        res.specs, ("paths", "/pets", "post", "responses", "default", "schema")
    )
    assert "$ref" in val

    val = path_get(res.specs, ("paths", "/pets/{petId}", "get", "parameters", 0))
    assert "$ref" in val

    val = path_get(
        res.specs, ("paths", "/pets/{petId}", "get", "responses", "200", "schema")
    )
    assert "$ref" not in val

    val = path_get(
        res.specs, ("paths", "/pets/{petId}", "get", "responses", "203", "schema")
    )
    assert "$ref" in val

    val = path_get(
        res.specs, ("paths", "/pets/{petId}", "get", "responses", "default", "schema")
    )
    assert "$ref" in val


def test_issue_23_partial_resolution_invalid_scheme():
    specs = {"$ref": "foo://cannot-do-anything"}
    res = resolver.RefResolver(specs, fs.abspath("tests/specs/with_externals.yaml"))

    with pytest.raises(ValueError):
        res.resolve_references()


def test_issue_69_urlparse_error():
    # XXX depending on python version, the error may not actually come from
    #     parsing the URL, but from trying to load a local file that does
    #     not exist. See test_issue_72_* for a test case that specifically
    #     tries for a nonexistent file url.
    specs = {"$ref": "file://a\u2100b/bad/netloc"}
    res = resolver.RefResolver(specs, fs.abspath("tests/specs/with_externals.yaml"))

    with pytest.raises(ResolutionError) as exinfo:
        res.resolve_references()

    assert "bad/netloc" in str(exinfo.value)


def test_issue_72_nonexistent_file_error():
    specs = {"$ref": "file://does/not/exist"}
    res = resolver.RefResolver(specs, fs.abspath("tests/specs/with_externals.yaml"))

    with pytest.raises(ResolutionError) as exinfo:
        res.resolve_references()

    assert "not/exist" in str(exinfo.value)


@pytest.mark.skipif(none_of("openapi-spec-validator"), reason="Missing backends")
def test_issue_78_resolve_internal_bug():
    specs = ""
    with open("tests/specs/issue_78/openapi.json") as fh:
        specs = fh.read()

    from prance.util import formats

    specs = formats.parse_spec(specs, "openapi.json")

    res = resolver.RefResolver(
        specs, fs.abspath("openapi.json"), resolve_types=resolver.RESOLVE_FILES
    )
    res.resolve_references()

    from prance.util.path import path_get

    val = path_get(res.specs, ("paths", "/endpoint", "post", "requestBody", "content"))

    # Reference to file is resolved
    assert "application/json" in val
    # Internal reference within file is NOT resolved
    assert "$ref" in val["application/json"]


@pytest.mark.skipif(none_of("openapi-spec-validator"), reason="Missing backends")
def test_issue_77_translate_external():
    specs = ""
    with open("tests/specs/issue_78/openapi.json") as fh:
        specs = fh.read()

    from prance.util import formats

    specs = formats.parse_spec(specs, "openapi.json")

    res = resolver.RefResolver(
        specs,
        fs.abspath("openapi.json"),
        resolve_types=resolver.RESOLVE_FILES,
        resolve_method=resolver.TRANSLATE_EXTERNAL,
    )
    res.resolve_references()

    from prance.util.path import path_get

    val = path_get(res.specs, ("components", "schemas", "_schemas.json_Body"))

    # Reference to file is translated in components/schemas
    assert "content" in val
    assert "application/json" in val["content"]

    # Reference url is updated
    val = path_get(res.specs, ("paths", "/endpoint", "post", "requestBody", "$ref"))
    assert val == "#/components/schemas/_schemas.json_Body"


@pytest.mark.skipif(none_of("openapi-spec-validator"), reason="Missing backends")
def test_issue_77_translate_external_refs_internal():
    specs = ""
    with open("tests/specs/issue_78/openapi.json") as fh:
        specs = fh.read()

    from prance.util import formats

    specs = formats.parse_spec(specs, "openapi.json")

    res = resolver.RefResolver(
        specs,
        fs.abspath("openapi.json"),
        resolve_types=resolver.RESOLVE_FILES | resolver.RESOLVE_INTERNAL,
        resolve_method=resolver.TRANSLATE_EXTERNAL,
    )
    res.resolve_references()

    from prance.util.path import path_get

    val = path_get(res.specs, ("components", "schemas", "_schemas.json_Body"))

    # Reference to file is translated in components/schemas
    assert "content" in val
    assert "application/json" in val["content"]

    # Internal Reference links updated
    assert (
        "#/components/schemas/_schemas.json_Something"
        == val["content"]["application/json"]["$ref"]
    )

    # Internal references is copied to components/schemas separately
    val = path_get(res.specs, ("components", "schemas", "_schemas.json_Something"))
    assert "type" in val

    # File reference url is updated as well
    val = path_get(res.specs, ("paths", "/endpoint", "post", "requestBody", "$ref"))
    assert val == "#/components/schemas/_schemas.json_Body"


@pytest.mark.skipif(none_of("openapi-spec-validator"), reason="Missing backends")
def test_issue_77_internal_refs_unresolved():
    specs = ""
    with open("tests/specs/issue_78/openapi.json") as fh:
        specs = fh.read()

    from prance.util import formats

    specs = formats.parse_spec(specs, "openapi.json")

    res = resolver.RefResolver(
        specs,
        fs.abspath("openapi.json"),
        resolve_types=resolver.RESOLVE_FILES,
        resolve_method=resolver.TRANSLATE_EXTERNAL,
    )
    res.resolve_references()

    from prance.util.path import path_get

    val = path_get(res.specs, ("components", "schemas"))

    # File reference resolved
    assert "_schemas.json_Body" in val

    # Internal file reference not resolved
    assert "_schemas.json_Something" not in val


@pytest.mark.xfail(reason="https://github.com/RonnyPfannschmidt/prance/issues/87")
def test_issue_205_swagger_resolution_failure():
    specs = ""
    with open("tests/specs/kubernetes_api_docs.json") as fh:
        specs = fh.read()

    from prance.util import formats

    specs = formats.parse_spec(specs, "kubernetes_api_docs.json")

    res = resolver.RefResolver(
        specs,
        fs.abspath("kubernetes_api_docs.json"),
        resolve_types=resolver.RESOLVE_FILES,
        resolve_method=resolver.TRANSLATE_EXTERNAL,
    )
    # test will raise an exception
    res.resolve_references()


@patch("requests.get")
def test_resolve_package_ref(mock_get):
    mock_get.side_effect = mock_get_petstore

    specs = get_specs("tests/specs/with_externals.yaml")
    res = resolver.RefResolver(
        specs,
        fs.abspath("tests/specs/with_externals.yaml"),
        resolve_types=resolver.RESOLVE_FILES,
    )
    res.resolve_references()

    from prance.util.path import path_get

    val = path_get(
        res.specs, ("paths", "/pets/{petId}", "get", "responses", "203", "schema")
    )
    assert "required" in val



"""Test suite for prance.util.iterators ."""

__author__ = "Jens Finkhaeuser"
__copyright__ = "Copyright (c) 2016-2021 Jens Finkhaeuser"
__license__ = "MIT"
__all__ = ()


from prance.util import iterators


def test_item_iterator():
    tester = {
        "foo": 42,
        "bar": {
            "some": "dict",
        },
        "baz": [
            {1: 2},
            {3: 4},
        ],
    }

    frozen = tuple(iterators.item_iterator(tester))
    assert len(frozen) == 9, "Iterator did not yield the correct number of items!"

    # Test for a few paths
    assert ((), tester) in frozen
    assert (("bar", "some"), "dict") in frozen
    assert (("baz", 0, 1), 2) in frozen
    assert (("baz", 1), {3: 4}) in frozen


def test_item_iterator_empty():
    frozen = tuple(iterators.item_iterator({}))
    assert len(frozen) == 1, "Must return at least the item itself!"
    assert ((), {}) in frozen

    frozen = tuple(iterators.item_iterator([]))
    assert len(frozen) == 1, "Must return at least the item itself!"
    assert ((), []) in frozen


def test_reference_iterator_empty_dict():
    tester = {}
    frozen = tuple(iterators.reference_iterator(tester))
    assert len(frozen) == 0, "Found items when it should not have!"


def test_reference_iterator_dict_without_references():
    tester = {
        "foo": 42,
        "bar": "baz",
        "baz": {
            "quux": {
                "fnord": False,
            },
        },
    }
    frozen = tuple(iterators.reference_iterator(tester))
    assert len(frozen) == 0, "Found items when it should not have!"


def test_reference_iterator_dict_with_references():
    tester = {
        "foo": 42,
        "bar": "baz",
        "$ref": "root",
        "baz": {
            "$ref": "branch",
            "quux": {
                "fnord": False,
                "$ref": "leaf",
            },
        },
    }
    frozen = tuple(iterators.reference_iterator(tester))
    assert len(frozen) == 3, "Found a different item count than expected!"

    # We have three references with their paths here, but we don't know which
    # order the tuple has them in. Let's go through them all:
    expectations = {
        0: "root",
        1: "branch",
        2: "leaf",
    }
    for key, value, path in iterators.reference_iterator(tester):
        assert value == expectations[len(path)]



"""Test suite for prance.util.url ."""

__author__ = "Jens Finkhaeuser"
__copyright__ = "Copyright (c) 2016-2021 Jens Finkhaeuser"
__license__ = "MIT"
__all__ = ()

import sys

import pytest

from prance.util import url

from . import platform


def test_absurl_http():
    test = "http://foo.bar/asdf/#lala/quux"
    res = url.absurl(test)
    assert res.geturl() == test


def test_absurl_http_fragment():
    base = "http://foo.bar/asdf/#lala/quux"
    test = "#another"
    res = url.absurl(test, base)
    assert res.scheme == "http"
    assert res.netloc == "foo.bar"
    assert res.path == "/asdf/"
    assert res.fragment == "another"


@pytest.mark.skipif(platform("win32"), reason="Skip on win32")
def test_absurl_file_posix():
    base = "file:///etc/passwd"
    test = "group"
    expect = "file:///etc/group"
    res = url.absurl(test, base)
    assert res.geturl() == expect


@pytest.mark.skipif(platform("!win32"), reason="Skip on !win32")
def test_absurl_file_win32():
    base = "file:///c:/windows/notepad.exe"
    test = "regedit.exe"
    expect = "file:///c:/windows/regedit.exe"
    res = url.absurl(test, base)
    assert res.geturl() == expect


@pytest.mark.skipif(platform("win32"), reason="Skip on win32")
def test_absurl_absfile_posix():
    test = "file:///etc/passwd"
    res = url.absurl(test)
    assert res.geturl() == test


@pytest.mark.skipif(platform("!win32"), reason="Skip on !win32")
def test_absurl_absfile_win32():
    test = "file:///c:/windows/notepad.exe"
    res = url.absurl(test)
    assert res.geturl() == test


def test_absurl_fragment():
    base = "file:///etc/passwd"
    test = "#frag"
    with pytest.raises(url.ResolutionError):
        url.absurl(test)

    res = url.absurl(test, base)
    assert res.geturl() == "file:///etc/passwd#frag"


def test_absurl_relfile():
    base = "http://foo.bar"
    test = "relative.file"
    with pytest.raises(url.ResolutionError):
        url.absurl(test)
    with pytest.raises(url.ResolutionError):
        url.absurl(test, base)


@pytest.mark.skipif(platform("win32"), reason="Skip on win32")
def test_absurl_paths_posix():
    base = "/etc/passwd"
    test = "group"
    expect = "file:///etc/group"

    with pytest.raises(url.ResolutionError):
        url.absurl(test)

    res = url.absurl(test, base)
    assert res.geturl() == expect


@pytest.mark.skipif(platform("!win32"), reason="Skip on !win32")
def test_absurl_paths_win32():
    base = "c:\\windows\\notepad.exe"
    test = "regedit.exe"
    expect = "file:///c:/windows/regedit.exe"

    with pytest.raises(url.ResolutionError):
        url.absurl(test)

    res = url.absurl(test, base)
    assert res.geturl() == expect


def test_urlresource():
    parsed = url.absurl("http://foo.bar/asdf?some=query#myfrag")
    res = url.urlresource(parsed)
    assert res == "http://foo.bar/asdf"


def test_split_url_reference():
    base = url.absurl("http://foo.bar/")

    # Relative reference
    parsed, path = url.split_url_reference(base, "#foo/bar")
    assert parsed.netloc == "foo.bar"
    assert len(path) == 2
    assert path[0] == "foo"
    assert path[1] == "bar"

    # Leading slashes are stripped
    parsed, path = url.split_url_reference(base, "#///foo/bar")
    assert parsed.netloc == "foo.bar"
    assert len(path) == 2
    assert path[0] == "foo"
    assert path[1] == "bar"

    # Absolute reference
    parsed, path = url.split_url_reference(base, "http://somewhere/#foo/bar")
    assert parsed.netloc == "somewhere"
    assert len(path) == 2
    assert path[0] == "foo"
    assert path[1] == "bar"

    # Reference with escaped parts
    parsed, path = url.split_url_reference(
        base, "http://somewhere/#foo/bar~1baz~1quux~0foo"
    )
    assert parsed.netloc == "somewhere"
    assert len(path) == 2
    assert path[0] == "foo"
    assert path[1] == "bar/baz/quux~foo"


def test_fetch_url_file():
    from prance.util import fs

    content = url.fetch_url(url.absurl(fs.abspath("tests/specs/with_externals.yaml")))
    assert content["swagger"] == "2.0"


def test_fetch_url_cached():
    from prance.util import fs

    cache = {}

    content1 = url.fetch_url(
        url.absurl(fs.abspath("tests/specs/with_externals.yaml")), cache
    )
    assert content1["swagger"] == "2.0"

    content2 = url.fetch_url(
        url.absurl(fs.abspath("tests/specs/with_externals.yaml")), cache
    )
    assert content2["swagger"] == "2.0"

    # Dicts are mutable, therefore we can't compare IDs. But individual
    # string fields should not be copied, because we shallow copy.
    assert id(content1["swagger"]) == id(content2["swagger"])


def test_fetch_url_text_cached():
    from prance.util import fs

    cache = {}

    content1, _ = url.fetch_url_text(
        url.absurl(fs.abspath("tests/specs/with_externals.yaml")), cache
    )
    content2, _ = url.fetch_url_text(
        url.absurl(fs.abspath("tests/specs/with_externals.yaml")), cache
    )

    # Strings are immutable, therefore IDs should be identical
    assert id(content1) == id(content2)


@pytest.mark.requires_network()
def test_fetch_url_http():
    exturl = "https://petstore.swagger.io/v2/swagger.yaml" "#/definitions/Pet"
    content = url.fetch_url(url.absurl(exturl))
    assert content["swagger"] == "2.0"


def test_fetch_url_python():
    exturl = "python://tests/specs/petstore.yaml"
    content = url.fetch_url(url.absurl(exturl))
    assert content["swagger"] == "2.0"



"""Test OpenAPI specs examples."""

__author__ = "Jens Finkhaeuser"
__copyright__ = "Copyright (c) 2018 Jens Finkhaeuser"
__license__ = "MIT"
__all__ = ()

import pytest

from prance import ResolvingParser
from prance import ValidationError
from prance.util.fs import FileNotFoundError


def make_name(path, parser, backend, version, file_format, entry):
    import os.path

    basename = os.path.splitext(entry)[0]
    basename = basename.replace("-", "_")
    version = version.replace(".", "")
    backend = backend.replace("-", "_")

    name = "_".join(["test", parser, backend, version, file_format, basename])
    return name


# Generate test cases at import.
# One case per combination of:
# - parser (base, resolving)
# - validation backend (flex, swagger-spec-validator) (skip ssv if importing it fails)
# - spec version
# - file format
# - file
# That gives >50 test cases

import os, os.path

base = os.path.join("tests", "examples")


def iter_entries(parser, backend, version, file_format, path):
    if version == "v3.0" and backend != "openapi-spec-validator":
        return

    for entry in os.listdir(path):
        full = os.path.join(path, entry)
        testcase_name = None
        if os.path.isfile(full):
            testcase_name = make_name(
                full, parser, backend, version, file_format, entry
            )
        elif os.path.isdir(full):
            if parser == "BaseParser":
                continue  # skip separate files for the BaseParser
            full = os.path.join(full, "spec", "swagger.%s" % (file_format))
            if os.path.isfile(full):
                testcase_name = make_name(
                    full, parser, backend, version, file_format, entry
                )

        if testcase_name:
            dirname = os.path.dirname(full)
            dirname = dirname.replace("\\", "\\\\")
            from prance.util import url

            absurl = url.absurl(os.path.abspath(full)).geturl()
            code = f"""
@pytest.mark.xfail()
def {testcase_name}():
  import os
  cur = os.getcwd()

  os.chdir('{dirname}')

  from prance import {parser}
  try:
    parser = {parser}('{absurl}', backend = '{backend}')
  finally:
    os.chdir(cur)
"""
            print(code)
            exec(code, globals())


for parser in ("BaseParser", "ResolvingParser"):
    from prance.util import validation_backends

    for backend in validation_backends():
        for version in os.listdir(base):
            version_dir = os.path.join(base, version)
            for file_format in os.listdir(version_dir):
                format_dir = os.path.join(version_dir, file_format)

                if not os.path.isdir(format_dir):  # Assume YAML
                    iter_entries(parser, backend, version, "yaml", version_dir)
                else:
                    for entry in os.listdir(format_dir):
                        iter_entries(parser, backend, version, file_format, format_dir)



from os.path import join
from re import match

import pytest

from . import none_of
from prance import _TranslatingParser


@pytest.fixture
def tester(request):
    pattern = r"test_(.+)"
    test_function_name = request.node.name
    name_match = match(pattern, test_function_name)
    test_name = name_match[1]

    file_name = f"{test_name}.spec.yaml"
    path = join("tests", "specs", "translating_parser", file_name)
    return SpecificationTester(path)


class SpecificationTester:
    """
    Helper class that tests correct reference translation. Loads a specification file from tests/specs, parses it with
    the TranslatingParser and provides assertion methods:
    * assert_schemas to check what’s in /components/schemas.
    * assert_path_ref to check the $ref of the only operation response in the specification.
    * assert_schema_ref to check the $ref of a schema in /components/schemas.

    Typical usage:
    tester = SpecificationTester("my_spec_file")
    tester.assert_path_ref("some_other_refd_file.spec.yaml_SomeObject")
    tester.assert_schemas(
        {"some_other_refd_file.spec.yaml_SomeObject", "some_other_refd_file.spec.yaml_SomeOtherObject"}
    )
    tester.assert_schema_ref(
        "some_other_refd_file.spec.yaml_SomeObject", "some_other_refd_file.spec.yaml_SomeOtherObject",
    )
    """

    @staticmethod
    def _assert_ref(schema, ref):
        assert "$ref" in schema
        assert schema["$ref"] == f"#/components/schemas/{ref}"

    def __init__(self, url):
        parser = _TranslatingParser(url, backend="openapi-spec-validator")
        parser.parse()
        self.specification = parser.specification

    def assert_schemas(self, keys):
        """
        Asserts schema names in /components/schemas.
        """
        assert self.specification["components"]["schemas"].keys() == keys

    def assert_path_ref(self, ref):
        """
        Asserts $ref value in /paths/hosts/get/responses/default/content/application/json/schema.
        """
        responses = self.specification["paths"]["/hosts"]["get"]["responses"]
        schema = responses["default"]["content"]["application/json"]["schema"]
        self._assert_ref(schema, ref)

    def assert_schema_ref(self, key, ref, getter=lambda schema: schema):
        """
        Asserts $ref value in /components/schemas/{key}. May use a custom getter function to find the $ref deeper in
        the schema.

        Example:
            tester.assert_schema_ref(
                "SomeComplexObject",
                "schemas.spec.yaml_SomeOtherObject",
                lambda schema: schema["properties"]["some_property"]
            )
        """
        schemas = self.specification["components"]["schemas"]
        assert key in schemas
        schema = getter(schemas[key])
        self._assert_ref(schema, ref)


@pytest.mark.skipif(
    none_of("openapi-spec-validator"), reason="Missing openapi-spec-validator"
)
def test_local_reference_from_root(tester):
    tester.assert_path_ref("PlainObject")
    tester.assert_schemas({"PlainObject"})


@pytest.mark.skipif(
    none_of("openapi-spec-validator"), reason="Missing openapi-spec-validator"
)
def test_file_reference_from_root(tester):
    tester.assert_path_ref("file_reference_from_root_schemas.spec.yaml_PlainObject")
    tester.assert_schemas({"file_reference_from_root_schemas.spec.yaml_PlainObject"})


@pytest.mark.skipif(
    none_of("openapi-spec-validator"), reason="Missing openapi-spec-validator"
)
def test_local_reference_from_file(tester):
    tester.assert_path_ref("local_reference_from_file_schemas.spec.yaml_RefObject")
    tester.assert_schemas(
        {
            "local_reference_from_file_schemas.spec.yaml_RefObject",
            "local_reference_from_file_schemas.spec.yaml_PlainObject",
        }
    )
    tester.assert_schema_ref(
        "local_reference_from_file_schemas.spec.yaml_RefObject",
        "local_reference_from_file_schemas.spec.yaml_PlainObject",
    )


@pytest.mark.skipif(
    none_of("openapi-spec-validator"), reason="Missing openapi-spec-validator"
)
def test_same_file_reference_from_file(tester):
    tester.assert_path_ref("same_file_reference_from_file_schemas.spec.yaml_RefObject")
    tester.assert_schemas(
        {
            "same_file_reference_from_file_schemas.spec.yaml_RefObject",
            "same_file_reference_from_file_schemas.spec.yaml_PlainObject",
        }
    )
    tester.assert_schema_ref(
        "same_file_reference_from_file_schemas.spec.yaml_RefObject",
        "same_file_reference_from_file_schemas.spec.yaml_PlainObject",
    )


@pytest.mark.skipif(
    none_of("openapi-spec-validator"), reason="Missing openapi-spec-validator"
)
def test_different_file_reference_from_file(tester):
    tester.assert_path_ref(
        "different_file_reference_from_file_schemas1.spec.yaml_RefObject"
    )
    tester.assert_schemas(
        {
            "different_file_reference_from_file_schemas1.spec.yaml_RefObject",
            "different_file_reference_from_file_schemas2.spec.yaml_PlainObject",
        }
    )
    tester.assert_schema_ref(
        "different_file_reference_from_file_schemas1.spec.yaml_RefObject",
        "different_file_reference_from_file_schemas2.spec.yaml_PlainObject",
    )


@pytest.mark.skipif(
    none_of("openapi-spec-validator"), reason="Missing openapi-spec-validator"
)
def test_root_file_reference_from_file(tester):
    tester.assert_path_ref("root_file_reference_from_file_schemas.spec.yaml_RefObject")
    tester.assert_schemas(
        {"PlainObject", "root_file_reference_from_file_schemas.spec.yaml_RefObject"}
    )
    tester.assert_schema_ref(
        "root_file_reference_from_file_schemas.spec.yaml_RefObject", "PlainObject"
    )


@pytest.mark.skipif(
    none_of("openapi-spec-validator"), reason="Missing openapi-spec-validator"
)
def test_root_file_reference_from_root(tester):
    tester.assert_path_ref("PlainObject")
    tester.assert_schemas({"PlainObject"})


@pytest.mark.skipif(
    none_of("openapi-spec-validator"), reason="Missing openapi-spec-validator"
)
def test_recursive_reference_in_root(tester):
    tester.assert_schema_ref(
        "RecursiveObject",
        "RecursiveObject",
        lambda schema: schema["additionalProperties"],
    )


@pytest.mark.skipif(
    none_of("openapi-spec-validator"), reason="Missing openapi-spec-validator"
)
def test_recursive_reference_in_file(tester):
    tester.assert_path_ref(
        "recursive_reference_in_file_schemas.spec.yaml_RecursiveObject"
    )
    tester.assert_schemas(
        {"recursive_reference_in_file_schemas.spec.yaml_RecursiveObject"}
    )
    tester.assert_schema_ref(
        "recursive_reference_in_file_schemas.spec.yaml_RecursiveObject",
        "recursive_reference_in_file_schemas.spec.yaml_RecursiveObject",
        lambda schema: schema["additionalProperties"],
    )


@pytest.mark.skipif(
    none_of("openapi-spec-validator"), reason="Missing openapi-spec-validator"
)
def test_nested_recursive_reference_in_file(tester):
    tester.assert_path_ref("Response")
    tester.assert_schemas(
        {
            "Response",
            "ResultsItem",
            "ReferenceObject",
            "nested_recursive_reference_in_file_schemas.spec.yaml_ComplexObject",
            "nested_recursive_reference_in_file_schemas.spec.yaml_ComplexObjectProperty",
            "nested_recursive_reference_in_file_schemas.spec.yaml_RecursiveObject",
        }
    )
    tester.assert_schema_ref(
        "nested_recursive_reference_in_file_schemas.spec.yaml_ComplexObject",
        "nested_recursive_reference_in_file_schemas.spec.yaml_ComplexObjectProperty",
        lambda schema: schema["properties"]["property"],
    )
    tester.assert_schema_ref(
        "nested_recursive_reference_in_file_schemas.spec.yaml_ComplexObjectProperty",
        "nested_recursive_reference_in_file_schemas.spec.yaml_RecursiveObject",
        lambda schema: schema["properties"]["recursive"],
    )
    tester.assert_schema_ref(
        "nested_recursive_reference_in_file_schemas.spec.yaml_RecursiveObject",
        "nested_recursive_reference_in_file_schemas.spec.yaml_RecursiveObject",
        lambda schema: schema["additionalProperties"]["oneOf"][0],
    )



"""Test suite focusing on validation backend features."""

__author__ = "Jens Finkhaeuser"
__copyright__ = "Copyright (c) 2018 Jens Finkhaeuser"
__license__ = "MIT"
__all__ = ()

import pytest

from prance import BaseParser
from prance import ValidationError
from prance.util import validation_backends

from . import none_of


def test_bad_backend():
    with pytest.raises(ValueError):
        BaseParser("tests/specs/petstore.yaml", backend="does_not_exist")


@pytest.mark.skipif(none_of("flex"), reason="Missing dependencies: flex")
def test_flex_issue_5_integer_keys():
    # Must succeed with default (flex) parser; note the parser does not stringify the response code
    parser = BaseParser("tests/specs/issue_5.yaml", backend="flex")
    assert 200 in parser.specification["paths"]["/test"]["post"]["responses"]


@pytest.mark.skipif(none_of("flex"), reason="Missing dependencies: flex")
def test_flex_validate_success():
    parser = BaseParser("tests/specs/petstore.yaml", backend="flex")


@pytest.mark.skipif(none_of("flex"), reason="Missing dependencies: flex")
def test_flex_validate_failure():
    with pytest.raises(ValidationError):
        parser = BaseParser("tests/specs/missing_reference.yaml", backend="flex")


@pytest.mark.skipif(
    none_of("swagger-spec-validator"),
    reason="Missing dependencies: swagger-spec-validator",
)
def test_swagger_spec_validator_issue_5_integer_keys():
    # Must fail in implicit strict mode.
    with pytest.raises(ValidationError):
        BaseParser("tests/specs/issue_5.yaml", backend="swagger-spec-validator")

    # Must fail in explicit strict mode.
    with pytest.raises(ValidationError):
        BaseParser(
            "tests/specs/issue_5.yaml", backend="swagger-spec-validator", strict=True
        )

    # Must succeed in non-strict/lenient mode
    parser = BaseParser(
        "tests/specs/issue_5.yaml", backend="swagger-spec-validator", strict=False
    )
    assert "200" in parser.specification["paths"]["/test"]["post"]["responses"]


@pytest.mark.skipif(
    none_of("swagger-spec-validator"),
    reason="Missing dependencies: swagger-spec-validator",
)
def test_swagger_spec_validator_validate_success():
    parser = BaseParser("tests/specs/petstore.yaml", backend="swagger-spec-validator")


@pytest.mark.skipif(
    none_of("swagger-spec-validator"),
    reason="Missing dependencies: swagger-spec-validator",
)
def test_swagger_spec_validator_validate_failure():
    with pytest.raises(ValidationError):
        parser = BaseParser(
            "tests/specs/missing_reference.yaml", backend="swagger-spec-validator"
        )


@pytest.mark.skipif(
    none_of("openapi-spec-validator"),
    reason="Missing dependencies: openapi-spec-validator",
)
def test_openapi_spec_validator_issue_5_integer_keys():
    # Must fail in implicit strict mode.
    with pytest.raises(ValidationError):
        BaseParser("tests/specs/issue_5.yaml", backend="openapi-spec-validator")

    # Must fail in explicit strict mode.
    with pytest.raises(ValidationError):
        BaseParser(
            "tests/specs/issue_5.yaml", backend="openapi-spec-validator", strict=True
        )

    # Must succeed in non-strict/lenient mode
    parser = BaseParser(
        "tests/specs/issue_5.yaml", backend="openapi-spec-validator", strict=False
    )
    assert "200" in parser.specification["paths"]["/test"]["post"]["responses"]


@pytest.mark.skipif(
    none_of("openapi-spec-validator"),
    reason="Missing dependencies: openapi-spec-validator",
)
def test_openapi_spec_validator_issue_36_error_reporting():
    with pytest.raises(ValidationError, match=r"Strict mode enabled"):
        BaseParser("tests/specs/issue_36.yaml", backend="openapi-spec-validator")


@pytest.mark.skipif(
    none_of("openapi-spec-validator"),
    reason="Missing dependencies: openapi-spec-validator",
)
def test_openapi_spec_validator_validate_success():
    parser = BaseParser("tests/specs/petstore.yaml", backend="openapi-spec-validator")


@pytest.mark.skipif(
    none_of("openapi-spec-validator"),
    reason="Missing dependencies: openapi-spec-validator",
)
def test_openapi_spec_validator_validate_failure():
    with pytest.raises(ValidationError):
        parser = BaseParser(
            "tests/specs/missing_reference.yaml", backend="openapi-spec-validator"
        )


@pytest.mark.skipif(
    none_of("openapi-spec-validator"),
    reason="Missing dependencies: openapi-spec-validator",
)
def test_openapi_spec_validator_issue_20_spec_version_handling():
    # The spec is OpenAPI 3, but broken. Need to set 'strict' to False to stringify keys
    with pytest.raises(ValidationError):
        parser = BaseParser(
            "tests/specs/issue_20.yaml", backend="openapi-spec-validator", strict=False
        )

    # Lazy parsing should let us validate what's happening
    parser = BaseParser(
        "tests/specs/issue_20.yaml",
        backend="openapi-spec-validator",
        strict=False,
        lazy=True,
    )
    assert not parser.valid
    assert parser.version_parsed == ()

    with pytest.raises(ValidationError):
        parser.parse()

    # After parsing, the specs are not valid, but the correct version is
    # detected.
    assert not parser.valid
    assert parser.version_parsed == (3, 0, 0)



"""Test suite for prance.BaseParser ."""

__author__ = "Jens Finkhaeuser"
__copyright__ = "Copyright (c) 2016-2021 Jens Finkhaeuser"
__license__ = "MIT"
__all__ = ()

import pytest

from prance import BaseParser
from prance import ValidationError
from prance.util.fs import FileNotFoundError

from . import none_of


@pytest.fixture
def petstore_parser():
    return BaseParser("tests/specs/petstore.yaml")


@pytest.fixture
def petstore_parser_from_string():
    yaml = None
    with open("tests/specs/petstore.yaml", "rb") as f:
        x = f.read()
        yaml = x.decode("utf8")
    return BaseParser(spec_string=yaml)


def test_load_fail():
    from prance.util.url import ResolutionError

    with pytest.raises(ResolutionError):
        BaseParser("tests/specs/missing.yaml")


def test_parse_fail():
    with pytest.raises(ValidationError):
        BaseParser(
            spec_string="""---
invalid 'a'sda YAML"""
        )


def test_version_fail():
    with pytest.raises(ValidationError):
        BaseParser(
            spec_string="""---
openapi: 4.0.0"""
        )


def test_filename_or_spec():
    with pytest.raises(AssertionError):
        BaseParser("", "")


@pytest.mark.skipif(
    none_of("openapi-spec-validator", "swagger-spec-validator", "flex"),
    reason="Backends missing.",
)
def test_load_and_parse_valid(petstore_parser):
    assert petstore_parser.specification, "No specs loaded!"


@pytest.mark.skipif(
    none_of("openapi-spec-validator", "swagger-spec-validator", "flex"),
    reason="Backends missing.",
)
def test_load_and_parse_lazy():
    parser = BaseParser("tests/specs/petstore.yaml", lazy=True)
    assert parser.specification is None, "Should not have specs yet!"

    parser.parse()
    assert parser.specification, "No specs loaded!"


@pytest.mark.skipif(
    none_of("openapi-spec-validator", "swagger-spec-validator", "flex"),
    reason="Backends missing.",
)
def test_yaml_valid(petstore_parser):
    assert petstore_parser.yaml(), "Did not get YAML representation of specs!"


@pytest.mark.skipif(
    none_of("openapi-spec-validator", "swagger-spec-validator", "flex"),
    reason="Backends missing.",
)
def test_json_valid(petstore_parser):
    assert petstore_parser.json(), "Did not get JSON representation of specs!"


@pytest.mark.skipif(
    none_of("openapi-spec-validator", "swagger-spec-validator", "flex"),
    reason="Backends missing.",
)
def test_cache_specs_mixin(petstore_parser):
    # In order to test the caching, we need to first use either the YAML or the
    # JSON mixin. Let's use YAML, because it's more swagger-ish
    yaml = petstore_parser.yaml()
    assert yaml, "Did not get YAML representation of specs!"

    # Caching should mean that if the specifications do not change, then neither
    # does the YAML representation.
    assert yaml == petstore_parser.yaml(), "YAML representation changed!"

    # In fact, the objects shouldn't even change.
    assert id(yaml) == id(petstore_parser.yaml()), (
        "YAML did not change but " "got regenerated!"
    )

    # However, when the specs change, then so must the YAML representation.
    petstore_parser.specification["foo"] = "bar"
    assert yaml != petstore_parser.yaml(), "YAML representation did not change!"


@pytest.mark.skipif(
    none_of("openapi-spec-validator", "swagger-spec-validator", "flex"),
    reason="Backends missing.",
)
def test_relative_urls_from_string(petstore_parser_from_string):
    # This must succeed
    assert (
        petstore_parser_from_string.yaml()
    ), "Did not get YAML representation of specs!"



"""Test suite prance.util.exceptions."""

__author__ = "Jens Finkhaeuser"
__copyright__ = "Copyright (c) 2018 Jens Finkhaeuser"
__license__ = "MIT"
__all__ = ()

import pytest

from prance.util import exceptions
from prance import ValidationError


def test_reraise_without_value_no_extra_message():
    with pytest.raises(ValidationError) as caught:
        exceptions.raise_from(ValidationError, None)

    # The first is obvious from pytest.raises. The rest tests
    # known attributes
    assert caught.type == ValidationError
    assert str(caught.value) == ""


def test_reraise_without_value_extra_message():
    with pytest.raises(ValidationError) as caught:
        exceptions.raise_from(ValidationError, None, "asdf")

    # The first is obvious from pytest.raises. The rest tests
    # known attributes
    assert caught.type == ValidationError
    assert str(caught.value) == "asdf"


def test_reraise_with_value_no_extra_message():
    with pytest.raises(ValidationError) as caught:
        try:
            raise RuntimeError("foo")
        except RuntimeError as inner:
            exceptions.raise_from(ValidationError, inner)

    # The first is obvious from pytest.raises. The rest tests
    # known attributes
    assert caught.type == ValidationError
    assert str(caught.value) == "foo"


def test_reraise_with_value_extra_message():
    with pytest.raises(ValidationError) as caught:
        try:
            raise RuntimeError("foo")
        except RuntimeError as inner:
            exceptions.raise_from(ValidationError, inner, "asdf")

    # The first is obvious from pytest.raises. The rest tests
    # known attributes
    assert caught.type == ValidationError
    assert str(caught.value) == "foo -- asdf"


def test_reraise_with_empty_value_string_extra_message():
    with pytest.raises(ValidationError) as caught:
        try:
            raise RuntimeError()
        except RuntimeError as inner:
            exceptions.raise_from(ValidationError, inner, "asdf")

    # The first is obvious from pytest.raises. The rest tests
    # known attributes
    assert caught.type == ValidationError
    assert str(caught.value) == "asdf"



"""Test suite for prance.util.path ."""

__author__ = "Jens Finkhaeuser"
__copyright__ = "Copyright (c) 2018 Jens Finkhaeuser"
__license__ = "MIT"
__all__ = ()


import pytest

from prance.util.path import path_set, path_get


def test_get_bad_path():
    # Raise with bad path types
    with pytest.raises(TypeError):
        path_get({}, 42)
    with pytest.raises(TypeError):
        path_get({}, 3.14)

    with pytest.raises(KeyError):
        path_get([], ("a", "b"))


def test_get_value_default():
    value = 42

    # No path can be resolved in a value type
    result = None
    with pytest.raises(TypeError):
        result = path_get(value, ("foo", "bar"), 123)
    assert result is None

    # However, we can resolve zero length paths
    result = path_get(value, (), 123)
    assert 42 == result
    result = path_get(None, (), 123)
    assert 123 == result

    # Also we can resolve None-type paths
    result = path_get(value, None, 321)
    assert 42 == result
    result = path_get(None, None, 321)
    assert 321 == result


def test_get_value_no_default():
    value = 42

    # No path can be resolved in a value type
    result = 666
    with pytest.raises(TypeError):
        result = path_get(value, ("foo", "bar"))
    assert result == 666

    # However, we can resolve zero length paths
    result = path_get(value, ())
    assert 42 == result
    result = path_get(None, ())
    assert result is None

    # Also we can resolve None-type paths
    result = path_get(value, None)
    assert 42 == result
    result = path_get(None, None)
    assert result is None


def test_get_collection_default():
    value = (1, 2, 3)

    # String paths in a Sequence should raise KeyError
    result = None
    with pytest.raises(KeyError):
        result = path_get(value, ("foo", "bar"), 123)
    assert result is None

    # A numeric path should work, though
    result = path_get(value, (1,), 123)
    assert 2 == result

    # Zero length paths should return the value or default value
    result = path_get(value, (), 123)
    assert (1, 2, 3) == result
    result = path_get(None, (), 123)
    assert 123 == result

    # And None paths as well
    result = path_get(value, None, 321)
    assert (1, 2, 3) == result
    result = path_get(None, None, 321)
    assert 321 == result


def test_get_collection_no_default():
    value = (1, 2, 3)

    # String paths in a Sequence should raise KeyError
    result = None
    with pytest.raises(KeyError):
        result = path_get(value, ("foo", "bar"))
    assert result is None

    # A numeric path should work, though
    result = path_get(value, (1,))
    assert 2 == result

    # Zero length paths should return the value or default value
    result = path_get(value, ())
    assert (1, 2, 3) == result
    result = path_get(None, ())
    assert result is None

    # And None paths as well
    result = path_get(value, None)
    assert (1, 2, 3) == result
    result = path_get(None, None)
    assert result is None


def test_get_mapping_default():
    value = {"foo": 1, "bar": 2, 3: 3}

    # String paths should work in a Mapping
    result = path_get(value, ("foo",), 123)
    assert 1 == result

    # So should numeric keys
    result = path_get(value, (3,), 123)
    assert 3 == result

    # Zero length paths should return the value or default value
    result = path_get(value, (), 123)
    assert {"foo": 1, "bar": 2, 3: 3} == result
    result = path_get(None, (), 123)
    assert 123 == result

    # And None paths as well
    result = path_get(value, None, 321)
    assert {"foo": 1, "bar": 2, 3: 3} == result
    result = path_get(None, None, 321)
    assert 321 == result


def test_set_bad_path():
    # Raise with bad path types
    with pytest.raises(TypeError):
        path_set({}, 42, None)
    with pytest.raises(TypeError):
        path_set({}, 3.14, None)

    with pytest.raises(KeyError):
        path_set([], ("a", "b"), None)


def test_set_value():
    value = 42

    # Setting on a value type with a path should raise...
    with pytest.raises(TypeError):
        path_set(42, ("foo", "bar"), "something")
    with pytest.raises(TypeError):
        path_set(42, (0, 1), "something")

    # ... and without a path or an empty path should raise an error
    with pytest.raises(KeyError):
        path_set(42, (), "something")
    with pytest.raises(TypeError):
        path_set(42, None, "something")


def test_set_sequence_no_create():
    # Tuples are not mutable - this must raise
    result = None
    with pytest.raises(TypeError):
        result = path_set((42,), (0,), "something")

    # Lists are mutable - this must work
    result = path_set(
        [
            42,
        ],
        (0,),
        "something",
    )
    assert ["something"] == result

    # Try nested paths
    result = path_set([42, [1, 2]], (1, 0), "something")
    assert [42, ["something", 2]] == result

    # Bad paths must raise
    with pytest.raises(IndexError):
        path_set([], (0,), "something")
    with pytest.raises(IndexError):
        path_set([[]], (0, 1), "something")


def test_set_sequence_create():
    # If we want to create entries, then mutable sequences should have further
    # sequences added to it.
    result = path_set([], (1, 0), "something", create=True)
    assert [None, ["something"]] == result


def test_set_mapping_no_create():
    # Setting existing keys must work
    result = path_set({"foo": "bar"}, ("foo",), "new")
    assert {"foo": "new"} == result

    # Nested mappings should work just as well.
    result = path_set(
        {"foo": {"bar": "baz"}},
        (
            "foo",
            "bar",
        ),
        "new",
    )
    assert {"foo": {"bar": "new"}} == result

    # Bad paths must raise
    with pytest.raises(KeyError):
        path_set({}, ("foo",), "something")
    with pytest.raises(KeyError):
        path_set(
            {"foo": {}},
            (
                "foo",
                "bar",
            ),
            "something",
        )


def test_set_mapping_create():
    # Setting missing keys must work
    result = path_set({}, ("foo",), "bar", create=True)
    assert {"foo": "bar"} == result

    # Recursive setting must work
    result = path_set(
        {},
        (
            "foo",
            "bar",
        ),
        "baz",
        create=True,
    )
    assert {"foo": {"bar": "baz"}} == result


def test_set_mixed_no_create():
    # Sequence in Mapping must work
    result = path_set({"foo": [0]}, ("foo", 0), 42)
    assert {"foo": [42]} == result

    # And Mapping in Sequence likewise
    result = path_set([{"foo": "bar"}], (0, "foo"), 42)
    assert [{"foo": 42}] == result


def test_set_mixed_create():
    # Creation should work based on the path element types
    result = path_set({}, ("foo", 0), 42, create=True)
    assert {"foo": [42]} == result

    result = path_set([], (0, "foo"), 42, create=True)
    assert [{"foo": 42}] == result

    # Create int keys in dict
    result = path_set({}, (0,), 42, create=True)
    assert {0: 42} == result


def test_set_mixed_create_no_fill():
    # Creation should not add items that are not necessary
    base = {"foo": [123]}

    result = path_set(base, ("foo", 0), 42, create=True)
    assert {"foo": [42]} == result


def test_get_informative_key_error():
    base = {"foo": {"bar": [123]}}

    # Match that the object being examining has its path printed, as
    # well as that the key is included.
    with pytest.raises(KeyError, match=r'.*"/".*asdf'):
        path_get(base, ("asdf",))

    with pytest.raises(KeyError, match=r'.*"/foo".*asdf'):
        path_get(base, ("foo", "asdf"))


def test_get_informative_index_error():
    base = {"foo": {"bar": [123]}}

    with pytest.raises(IndexError, match=r'.*"/foo/bar".*123'):
        path_get(base, ("foo", "bar", 123))



"""Test suite for prance.util.formats ."""

__author__ = "Jens Finkhaeuser"
__copyright__ = "Copyright (c) 2016-2021 Jens Finkhaeuser"
__license__ = "MIT"
__all__ = ()


import pytest

from prance.util import formats


def test_format_info_yaml():
    ctype, ext = formats.format_info("yaml")
    assert "yaml" in ctype
    assert "yaml" in ext


def test_format_info_json():
    ctype, ext = formats.format_info("json")
    assert "json" in ctype
    assert "json" in ext


def test_format_info_bad_format():
    ctype, ext = formats.format_info("foo")
    assert ctype is None
    assert ext is None


def test_parse_details_yaml():
    yaml = """---
foo: bar
"""
    parsed, ctype, ext = formats.parse_spec_details(yaml, "foo.yaml")

    assert parsed["foo"] == "bar", "Did not parse with explicit YAML"
    assert "yaml" in ctype
    assert "yaml" in ext


def test_parse_yaml():
    yaml = """---
foo: bar
"""
    parsed = formats.parse_spec(yaml, "foo.yaml")
    assert parsed["foo"] == "bar", "Did not parse with explicit YAML"

    parsed = formats.parse_spec(yaml)
    assert parsed["foo"] == "bar", "Did not parse with implicit YAML"


def test_parse_yaml_ctype():
    yaml = """---
foo: bar
"""
    parsed = formats.parse_spec(yaml, None, content_type="text/yaml")
    assert parsed["foo"] == "bar", "Did not parse with explicit YAML"


def test_parse_json():
    json = '{ "foo": "bar" }'

    parsed = formats.parse_spec(json, "foo.js")
    assert parsed["foo"] == "bar", "Did not parse with explicit JSON"


def test_parse_json_ctype():
    json = '{ "foo": "bar" }'

    parsed = formats.parse_spec(json, None, content_type="application/json")
    assert parsed["foo"] == "bar", "Did not parse with explicit JSON"


def test_parse_unknown():
    with pytest.raises(formats.ParseError):
        formats.parse_spec("{-")


def test_parse_unknown_ext():
    with pytest.raises(formats.ParseError):
        formats.parse_spec("{-", "asdf.xml")


def test_parse_unknown_ctype():
    with pytest.raises(formats.ParseError):
        formats.parse_spec("{-", None, content_type="text/xml")


def test_serialize_json():
    specs = {
        "foo": "bar",
        "baz": [1, 2, 3],
    }

    # With no further information given, the specs must be in JSON
    serialized = formats.serialize_spec(specs)
    assert serialized.startswith("{")

    # The same must be the case if we provide a JSON file name
    serialized = formats.serialize_spec(specs, "foo.json")
    assert serialized.startswith("{")


def test_serialize_yaml():
    specs = {
        "foo": "bar",
        "baz": [1, 2, 3],
    }

    # Provide a YAML file name
    serialized = formats.serialize_spec(specs, "foo.yml")
    assert "foo: bar" in serialized


def test_serialize_json_ctype():
    specs = {
        "foo": "bar",
        "baz": [1, 2, 3],
    }

    # Force JSON with content type
    serialized = formats.serialize_spec(specs, None, content_type="application/json")
    assert serialized.startswith("{")


def test_serialize_yaml_ctype():
    specs = {
        "foo": "bar",
        "baz": [1, 2, 3],
    }

    # Force YAML with content type
    serialized = formats.serialize_spec(specs, None, content_type="text/yaml")
    assert "foo: bar" in serialized



"""Test suite for prance.util ."""

__author__ = "Jens Finkhaeuser"
__copyright__ = "Copyright (c) 2016-2021 Jens Finkhaeuser"
__license__ = "MIT"
__all__ = ()


import pytest

from prance import util


def test_stringify_keys():
    test = {
        "str": 42,
        "unicode": 123,
        "nested": {
            "str": 42,
            "unicode": 123,
        },
        42: "int",
        3.14: "float",
    }

    result = util.stringify_keys(test)

    def assert_object_identity(key, first, second):
        ret = None
        for k1, v1 in first.items():
            for k2, v2 in second.items():
                if k1 == key and k2 == key:
                    assert id(k1) == id(k2)
                    ret = id(k1)
        assert ret is not None
        return ret

    # The number of keys must not have changed
    assert len(test) == len(result)
    assert len(test["nested"]) == len(result["nested"])

    # str/unicode keys must remain absolutely unchanged. We're testing for
    # object identity.
    found = []
    for key in ("str", "unicode"):
        found.append(assert_object_identity(key, test, result))
        found.append(assert_object_identity(key, test["nested"], result["nested"]))

    # for all other keys we find in the test set, there must exist a
    # stringified version of it in the result set
    for key, value in test.items():
        if key in found:
            continue

        assert str(key) in result

    for key, value in test["nested"].items():
        if key in found:
            continue

        assert str(key) in result["nested"]



"""Test suite for prance.cli ."""

__author__ = "Jens Finkhaeuser"
__copyright__ = "Copyright (c) 2016-2021 Jens Finkhaeuser"
__license__ = "MIT"
__all__ = ()

import pytest

from . import none_of


@pytest.fixture
def runner():
    from click.testing import CliRunner

    return CliRunner()


@pytest.mark.skipif(none_of("click"), reason="Click does not exist")
def test_validate_defaults(runner):
    from prance import cli

    # Good example
    result = runner.invoke(cli.validate, ["tests/specs/petstore.yaml"])
    assert result.exit_code == 0
    expected = """Processing "tests/specs/petstore.yaml"...
 -> Resolving external references.
Validates OK as Swagger/OpenAPI 2.0!
"""
    assert result.output == expected

    # Bad example
    result = runner.invoke(cli.validate, ["tests/specs/definitions.yaml"])
    assert result.exit_code == 1
    assert "ValidationError" in result.output


@pytest.mark.skipif(none_of("click"), reason="Click does not exist")
def test_validate_multiple(runner):
    from prance import cli

    result = runner.invoke(
        cli.validate, ["tests/specs/petstore.yaml", "tests/specs/petstore.yaml"]
    )
    assert result.exit_code == 0
    expected = """Processing "tests/specs/petstore.yaml"...
 -> Resolving external references.
Validates OK as Swagger/OpenAPI 2.0!
Processing "tests/specs/petstore.yaml"...
 -> Resolving external references.
Validates OK as Swagger/OpenAPI 2.0!
"""
    assert result.output == expected


@pytest.mark.skipif(none_of("click"), reason="Click does not exist")
def test_validate_no_resolve(runner):
    from prance import cli

    # Good example
    result = runner.invoke(cli.validate, ["--no-resolve", "tests/specs/petstore.yaml"])
    assert result.exit_code == 0
    expected = """Processing "tests/specs/petstore.yaml"...
 -> Not resolving external references.
Validates OK as Swagger/OpenAPI 2.0!
"""
    assert result.output == expected


@pytest.mark.skipif(none_of("click"), reason="Click does not exist")
def test_validate_output_too_many_inputs(runner):
    from prance import cli

    result = runner.invoke(
        cli.validate,
        ["-o", "foo", "tests/specs/petstore.yaml", "tests/specs/petstore.yaml"],
    )
    assert result.exit_code == 2
    assert "If --output-file is given," in result.output


@pytest.mark.skipif(none_of("click"), reason="Click does not exist")
def test_validate_output(runner):
    from prance import cli

    import os, os.path

    curdir = os.getcwd()

    outnames = ["foo.json", "foo.yaml"]
    for outname in outnames:
        with runner.isolated_filesystem():
            result = runner.invoke(
                cli.validate,
                ["-o", outname, os.path.join(curdir, "tests/specs/petstore.yaml")],
            )
            assert result.exit_code == 0

            # There also must be a 'foo' file now.
            files = [f for f in os.listdir(".") if os.path.isfile(f)]
            assert outname in files

            # Ensure a UTF-8 file encoding.
            from prance.util import fs

            assert "utf-8" in fs.detect_encoding(
                outname, default_to_utf8=False, read_all=True
            )

            # Now do a full encoding detection, too

            # The 'foo' file must be a valid swagger spec.
            result = runner.invoke(cli.validate, [outname])
            assert result.exit_code == 0


@pytest.mark.skipif(none_of("click"), reason="Click does not exist")
def test_compile_defaults(runner):
    from prance import cli

    # Good example
    result = runner.invoke(cli.compile, ["tests/specs/petstore.yaml"])
    assert result.exit_code == 0
    expected = """Processing "tests/specs/petstore.yaml"...
 -> Resolving external references.
Validates OK as Swagger/OpenAPI 2.0!
"""
    assert result.output.startswith(expected)

    # Bad example
    result = runner.invoke(cli.validate, ["tests/specs/definitions.yaml"])
    assert result.exit_code == 1
    assert "ValidationError" in result.output


@pytest.mark.skipif(none_of("click"), reason="Click does not exist")
def test_compile_output(runner):
    from prance import cli

    import os, os.path

    curdir = os.getcwd()

    outnames = ["foo.json", "foo.yaml"]
    for outname in outnames:
        with runner.isolated_filesystem():
            result = runner.invoke(
                cli.compile,
                [os.path.join(curdir, "tests/specs/petstore.yaml"), outname],
            )
            assert result.exit_code == 0

            # There also must be a 'foo' file now.
            files = [f for f in os.listdir(".") if os.path.isfile(f)]
            assert outname in files

            # Ensure a UTF-8 file encoding.
            from prance.util import fs

            assert "utf-8" in fs.detect_encoding(
                outname, default_to_utf8=False, read_all=True
            )

            # Now do a full encoding detection, too

            # The 'foo' file must be a valid swagger spec.
            result = runner.invoke(cli.validate, [outname])
            assert result.exit_code == 0


@pytest.mark.skipif(none_of("click"), reason="Click does not exist")
@pytest.mark.requires_network()
def test_convert_defaults(runner):
    from prance import cli

    # Good example
    result = runner.invoke(cli.convert, ["tests/specs/petstore.yaml"])
    assert result.exit_code == 0
    assert "openapi" in result.output
    assert "3." in result.output

    # Bad example
    result = runner.invoke(cli.validate, ["tests/specs/definitions.yaml"])
    assert result.exit_code == 1
    assert "ValidationError" in result.output


@pytest.mark.skipif(none_of("click"), reason="Click does not exist")
@pytest.mark.requires_network()
def test_convert_output(runner):
    from prance import cli

    import os, os.path

    curdir = os.getcwd()

    outnames = ["foo.json", "foo.yaml"]
    for outname in outnames:
        with runner.isolated_filesystem():
            result = runner.invoke(
                cli.convert,
                [os.path.join(curdir, "tests/specs/petstore.yaml"), outname],
            )
            assert result.exit_code == 0

            # There also must be a 'foo' file now.
            files = [f for f in os.listdir(".") if os.path.isfile(f)]
            assert outname in files

            # Ensure a UTF-8 file encoding.
            from prance.util import fs

            assert "utf-8" in fs.detect_encoding(
                outname, default_to_utf8=False, read_all=True
            )

            # Now do a full encoding detection, too
            contents = fs.read_file(outname)
            assert "openapi" in contents
            assert "3." in contents



"""Test suite for prance.ResolvingParser ."""

__author__ = "Jens Finkhaeuser"
__copyright__ = "Copyright (c) 2016-2021 Jens Finkhaeuser"
__license__ = "MIT"
__all__ = ()

import pytest
from unittest.mock import patch

from prance import ResolvingParser
from prance import ValidationError

from . import none_of


def mock_get_petstore(*args, **kwargs):
    from .mock_response import MockResponse, PETSTORE_YAML

    return MockResponse(text=PETSTORE_YAML)


@pytest.fixture
def petstore_parser():
    return ResolvingParser("tests/specs/petstore.yaml")


@pytest.fixture
@patch("requests.get")
def with_externals_parser(mock_get):
    mock_get.side_effect = mock_get_petstore
    return ResolvingParser("tests/specs/with_externals.yaml")


@pytest.fixture
def petstore_parser_from_string():
    yaml = None
    with open("tests/specs/petstore.yaml", "rb") as f:
        x = f.read()
        yaml = x.decode("utf8")
    return ResolvingParser(spec_string=yaml)


@pytest.fixture
def issue_1_parser():
    return ResolvingParser("tests/specs/issue_1.json")


@pytest.mark.skipif(
    none_of("openapi-spec-validator", "swagger-spec-validator", "flex"),
    reason="Missing backends",
)
def test_basics(petstore_parser):
    assert petstore_parser.specification, "No specs loaded!"


@pytest.mark.skipif(
    none_of("openapi-spec-validator", "swagger-spec-validator", "flex"),
    reason="Missing backends",
)
def test_petstore_resolve(petstore_parser):
    assert petstore_parser.specification, "No specs loaded!"

    # The petstore references /definitions/Pet in /definitions/Pets, and uses
    # /definitions/Pets in the 200 response to the /pets path. So let's check
    # whether we can find something of /definitions/Pet there...
    res = petstore_parser.specification["paths"]["/pets"]["get"]["responses"]
    assert res["200"]["schema"]["type"] == "array", "Did not resolve right!"


@pytest.mark.skipif(
    none_of("openapi-spec-validator", "swagger-spec-validator", "flex"),
    reason="Missing backends",
)
def test_with_externals_resolve(with_externals_parser):
    assert with_externals_parser.specification, "No specs loaded!"

    # The specs are a simplified version of the petstore example, with some
    # external references.
    # - Test that the list pets call returns the right thing from the external
    #   definitions.yaml
    res = with_externals_parser.specification["paths"]["/pets"]["get"]
    res = res["responses"]
    assert res["200"]["schema"]["type"] == "array"

    # - Test that the get single pet call returns the right thing from the
    #   remote petstore definition
    res = with_externals_parser.specification["paths"]["/pets/{petId}"]["get"]
    res = res["responses"]
    assert "id" in res["200"]["schema"]["required"]

    # - Test that error responses contain a message from error.json
    res = with_externals_parser.specification["paths"]["/pets"]["get"]
    res = res["responses"]
    assert "message" in res["default"]["schema"]["required"]


@pytest.mark.skipif(
    none_of("openapi-spec-validator", "swagger-spec-validator", "flex"),
    reason="Missing backends",
)
def test_relative_urls_from_string(petstore_parser_from_string):
    # This must succeed
    assert (
        petstore_parser_from_string.yaml()
    ), "Did not get YAML representation of specs!"


@pytest.mark.skipif(
    none_of("openapi-spec-validator", "swagger-spec-validator", "flex"),
    reason="Missing backends",
)
def test_issue_1_relative_path_references(issue_1_parser):
    # Must resolve references correctly
    params = issue_1_parser.specification["paths"]["/test"]["parameters"]
    assert "id" in params[0]["schema"]["required"]


@pytest.mark.skipif(none_of("openapi-spec-validator"), reason="Missing backends")
def test_issue_39_sequence_indices():
    # Must not fail to parse
    parser = ResolvingParser(
        "tests/specs/issue_39.yaml", backend="openapi-spec-validator"
    )

    # The /useCase path should have two values in its response example.
    example = parser.specification["paths"]["/useCase"]["get"]["responses"]["200"][
        "content"
    ]["application/json"]["examples"]["response"]
    assert "value" in example
    assert len(example["value"]) == 2

    # However, the /test path should have only one of the strings.
    example = parser.specification["paths"]["/test"]["get"]["responses"]["200"][
        "content"
    ]["application/json"]["example"]
    assert example == "some really long or specific string"


@pytest.mark.skipif(none_of("openapi-spec-validator"), reason="Missing backends")
def test_issue_51_encoding_error():
    # Parsing used to throw - but shouldn't after heuristic change.
    parser = ResolvingParser(
        "tests/specs/issue_51/openapi-main.yaml",
        lazy=True,
        backend="openapi-spec-validator",
        strict=True,
    )

    parser.parse()

    # Parsing with setting an explicit and wrong file encoding should raise
    # an error, effectively reverting to the old behaviour
    parser = ResolvingParser(
        "tests/specs/issue_51/openapi-main.yaml",
        lazy=True,
        backend="openapi-spec-validator",
        strict=True,
        encoding="iso-8859-2",
    )

    from ruamel.yaml.reader import ReaderError

    with pytest.raises(ReaderError):
        parser.parse()


@pytest.mark.skipif(none_of("openapi-spec-validator"), reason="Missing backends")
def test_issue_65_partial_resolution_files():
    specs = """openapi: "3.0.0"
info:
  title: ''
  version: '1.0.0'
paths: {}
components:
    schemas:
        SampleArray:
            type: array
            items:
              $ref: '#/components/schemas/ItemType'

        ItemType:
          type: integer
"""

    from prance.util import resolver

    parser = ResolvingParser(
        spec_string=specs,
        resolve_types=resolver.RESOLVE_FILES,
        backend="openapi-spec-validator",
    )

    from prance.util.path import path_get

    val = path_get(
        parser.specification, ("components", "schemas", "SampleArray", "items")
    )
    assert "$ref" in val


@pytest.mark.skipif(none_of("openapi-spec-validator"), reason="Missing backends")
def test_issue_83_skip_propagation():
    # Throw with strict parsing
    parser = ResolvingParser(
        "tests/specs/issue_83/bad_spec.yml",
        lazy=True,
        backend="openapi-spec-validator",
        strict=True,
    )

    with pytest.raises(ValidationError):
        parser.parse()

    # Do not throw with non-strict parsing
    parser = ResolvingParser(
        "tests/specs/issue_83/bad_spec.yml",
        lazy=True,
        backend="openapi-spec-validator",
        strict=False,
    )

    parser.parse()


@pytest.mark.skipif(none_of("openapi-spec-validator"), reason="Missing backends")
def test_value_not_converted_to_boolean():
    specs = """openapi: "3.0.0"
info:
  title: ''
  version: '1.0.0'
paths: {}
components:
    schemas:
        SampleEnum:
            type: string
            enum:
              - NO
              - OFF
"""

    from prance.util import resolver

    parser = ResolvingParser(
        spec_string=specs,
        resolve_types=resolver.RESOLVE_FILES,
        backend="openapi-spec-validator",
    )
    specs = parser.specification
    assert specs["components"]["schemas"]["SampleEnum"]["enum"] == ["NO", "OFF"]



"""Test suite for prance.convert ."""

__author__ = "Jens Finkhaeuser"
__copyright__ = "Copyright (c) 2018-2021 Jens Finkhaeuser"
__license__ = "MIT"
__all__ = ()

import pytest

from . import none_of

from prance import convert


@pytest.fixture
def petstore_yaml():
    from prance.util import fs

    return fs.read_file("tests/examples/v2.0/yaml/petstore.yaml")


@pytest.fixture
def petstore_json():
    from prance.util import fs

    return fs.read_file("tests/examples/v2.0/json/petstore.json")


@pytest.mark.requires_network()
def test_convert_petstore_yaml(petstore_yaml):
    converted, content_type = convert.convert_str(
        petstore_yaml, content_type="application/yaml"
    )

    # Check correct content type
    assert "yaml" in content_type

    # Parsing can't fail.
    from prance.util import formats

    parsed = formats.parse_spec(converted, content_type=content_type)

    # Assert the correct target version
    assert "openapi" in parsed
    assert parsed["openapi"].startswith("3.")


@pytest.mark.requires_network()
def test_convert_petstore_json(petstore_json):
    converted, content_type = convert.convert_str(
        petstore_json, content_type="application/json"
    )

    # Check correct content type
    assert "json" in content_type

    # Parsing can't fail.
    from prance.util import formats

    parsed = formats.parse_spec(converted, content_type=content_type)

    # Assert the correct target version
    assert "openapi" in parsed
    assert parsed["openapi"].startswith("3.")


@pytest.mark.requires_network()
def test_convert_petstore_yaml_explicit_name(petstore_yaml):
    converted, content_type = convert.convert_str(petstore_yaml, filename="foo.yml")

    # Check correct content type
    assert "yaml" in content_type


@pytest.mark.requires_network()
def test_convert_url():
    from prance.util import url

    converted, content_type = convert.convert_url(
        url.absurl("python://tests/specs/petstore.yaml")
    )

    # Check correct content type
    assert "yaml" in content_type

    # Parsing can't fail.
    from prance.util import formats

    parsed = formats.parse_spec(converted, content_type=content_type)

    # Assert the correct target version
    assert "openapi" in parsed
    assert parsed["openapi"].startswith("3.")


@pytest.mark.requires_network()
def test_convert_spec():
    from prance import BaseParser, ResolvingParser, ValidationError

    parser = BaseParser("tests/specs/petstore.yaml")

    # Conversion should fail with the default backend.
    with pytest.raises(ValidationError):
        converted = convert.convert_spec(parser.specification, backend="flex")

    # However, with the lazy flag it should work.
    converted = convert.convert_spec(parser.specification, lazy=True)
    assert isinstance(converted, BaseParser)

    # Passing a ResolvingParser class should also work.
    converted = convert.convert_spec(parser.specification, ResolvingParser, lazy=True)
    assert isinstance(converted, ResolvingParser)


@pytest.mark.requires_network()
def test_convert_parser_lazy_swagger_backend():
    from prance import BaseParser, ResolvingParser, ValidationError

    parser = BaseParser("tests/specs/petstore.yaml")

    # Conversion should fail with the default backend.
    with pytest.raises(ValidationError):
        converted = convert.convert_spec(parser, backend="flex")

    # However, with the lazy flag it should work.
    converted = convert.convert_spec(parser, lazy=True)
    assert isinstance(converted, BaseParser)

    # Passing a ResolvingParser class should also work.
    converted = convert.convert_spec(parser, ResolvingParser, lazy=True)
    assert isinstance(converted, ResolvingParser)


@pytest.mark.skipif(
    none_of("openapi-spec-validator"), reason="Missing openapi-spec-validator"
)
@pytest.mark.requires_network()
def test_convert_parser_validated():
    from prance import BaseParser

    parser = BaseParser("tests/specs/petstore.yaml", backend="openapi-spec-validator")

    # Conversion should work: it's the right backend, and it validates.
    converted = convert.convert_spec(parser)
    assert isinstance(converted, BaseParser)
    assert converted.version_parsed[0] == 3



"""Utility code for tests."""

__author__ = "Jens Finkhaeuser"
__copyright__ = "Copyright (c) 2017-2021 Jens Finkhaeuser"
__license__ = "MIT"
__all__ = ()

import pytest


class Sandboxer:
    """
    Given a sandbox template path, uses pytest to create a temporary directory,
    then copies the template's contents there, and changes to the directory.

    At exit, it changes back to the current working directory.
    """

    def __init__(self, tmpdir, path=None):
        self.__tmpdir = tmpdir
        self.__templatepath = path

    def __enter__(self):
        # Save current working directory & python path
        import os, sys

        self.__cwd = os.getcwd()
        self.__syspath = sys.path[:]

        # Create a new temporary directory and copy the template over
        if self.__templatepath:
            tmp = str(self.__tmpdir.join(self.__templatepath))
            import shutil

            shutil.copytree(self.__templatepath, tmp)
        else:
            tmp = self.__tmpdir

        # Then change to the temporary directory & python path
        os.chdir(str(tmp))
        import site

        sys.path.insert(0, tmp)

    def __exit__(self, exc_type, exc_value, traceback):
        # Restore current working directory & python path
        import os, sys

        os.chdir(str(self.__cwd))
        sys.path = self.__syspath

        # Don't suppress exceptions
        return False


def sandbox(*args, **kwargs):
    """Create a sandbox"""
    return Sandboxer(*args, **kwargs)



"""Test suite for prance.util.fs ."""

__author__ = "Jens Finkhaeuser"
__copyright__ = "Copyright (c) 2016-2021 Jens Finkhaeuser"
__license__ = "MIT"
__all__ = ()


import os
import sys

import pytest

from prance.util import fs

from . import sandbox, platform


@pytest.fixture
def create_symlink():
    # Defined as a fixture so we can selectively use it for tests."
    testname = "tests/specs/symlink_test"
    import os.path

    if os.path.islink(testname):
        return testname

    if os.path.exists(testname):
        os.unlink(testname)

    target = "with_externals.yaml"
    os.symlink(target, testname)

    return testname


@pytest.mark.skipif(platform("win32"), reason="Skip on win32")
def test_canonical(create_symlink):
    res = fs.canonical_filename(create_symlink)
    expected = os.path.join(os.getcwd(), "tests/specs/with_externals.yaml")
    assert res == expected


def test_to_posix_rel():
    test = "tests/specs/with_externals.yaml"
    assert fs.to_posix(os.path.normpath(test)) == test


@pytest.mark.skipif(platform("win32"), reason="Skip on win32")
def test_to_posix_abs_posix():
    test = "/etc/passwd"
    expected = test
    assert fs.to_posix(test) == expected


@pytest.mark.skipif(platform("!win32"), reason="Skip on !win32")
def test_to_posix_abs_win32():
    test = "c:\\windows\\notepad.exe"
    expected = "/c:/windows/notepad.exe"
    assert fs.to_posix(test) == expected


def test_from_posix_rel():
    test = "tests/specs/with_externals.yaml"
    assert fs.from_posix(test) == os.path.normpath(test)


@pytest.mark.skipif(platform("win32"), reason="Skip on win32")
def test_from_posix_abs_posix():
    test = "/etc/passwd"
    expected = test
    assert fs.from_posix(test) == expected


@pytest.mark.skipif(platform("!win32"), reason="Skip on !win32")
def test_from_posix_abs_win32():
    test = "/c:/windows/notepad.exe"
    expected = "c:\\windows\\notepad.exe"
    assert fs.from_posix(test) == expected


def test_abspath_basics():
    testname = os.path.normpath("tests/specs/with_externals.yaml")
    res = fs.abspath(testname)
    expected = fs.to_posix(os.path.join(os.getcwd(), testname))
    assert res == expected


def test_abspath_relative():
    testname = "error.json"
    relative = os.path.join(os.getcwd(), "tests/specs/with_externals.yaml")
    res = fs.abspath(testname, relative)
    expected = fs.to_posix(os.path.join(os.getcwd(), "tests", "specs", testname))
    assert res == expected


def test_abspath_relative_dir():
    testname = "error.json"
    relative = os.path.join(os.getcwd(), "tests", "specs")
    res = fs.abspath(testname, relative)
    expected = fs.to_posix(os.path.join(os.getcwd(), "tests", "specs", testname))
    assert res == expected


def test_detect_encoding():
    # Quick detection should yield utf-8 for the petstore file.
    assert fs.detect_encoding("tests/specs/petstore.yaml") == "utf-8"

    # Without defaulting to utf-8, it should also work.
    assert (
        fs.detect_encoding("tests/specs/petstore.yaml", default_to_utf8=False)
        == "utf-8"
    )

    # Deep inspection should not change anything because the file size is
    # too small.
    assert (
        fs.detect_encoding(
            "tests/specs/petstore.yaml", default_to_utf8=False, read_all=True
        )
        == "utf-8"
    )

    # The UTF-8 file with BOM should be detected properly
    assert fs.detect_encoding("tests/specs/utf8bom.yaml") == "utf-8-sig"


def test_issue_51_detect_encoding():
    # This should be UTF-8, but in issue 51 it was reported that it comes back
    # as iso-8859-2

    # Detection should be ok if the entire file is read
    assert (
        fs.detect_encoding("tests/specs/issue_51/openapi-part.yaml", read_all=True)
        == "utf-8"
    )

    # After the heuristic change, it reads the whole file anyway.
    assert fs.detect_encoding("tests/specs/issue_51/openapi-part.yaml") == "utf-8"

    # Specifically re-encoded as iso-8859-2 should fail - but not as
    # a call to the detect_encoding() function. Instead, we can only return
    # a badly detected encoding. Chardet sends iso-8859-1 here.
    assert (
        fs.detect_encoding("tests/specs/issue_51/openapi-part-iso-8859-2.yaml")
        == "iso-8859-1"
    )


def test_load_nobom():
    contents = fs.read_file("tests/specs/petstore.yaml")
    assert contents.index("Swagger Petstore") >= 0, "File reading failed!"


def test_load_utf8bom():
    contents = fs.read_file("tests/specs/utf8bom.yaml")
    assert contents.index("söme välüe") >= 0, "UTF-8 BOM handling failed!"


def test_load_utf8bom_override():
    with pytest.raises(UnicodeDecodeError):
        fs.read_file("tests/specs/utf8bom.yaml", "ascii")


def test_write_file(tmpdir):
    with sandbox.sandbox(tmpdir):
        test_text = "söme täxt"
        fs.write_file("test.out", test_text)

        # File must have been written
        files = [f for f in os.listdir(".") if os.path.isfile(f)]
        assert "test.out" in files

        # File contents must work
        contents = fs.read_file("test.out")
        assert test_text == contents


def test_write_file_bom(tmpdir):
    with sandbox.sandbox(tmpdir):
        test_text = "söme täxt"
        fs.write_file("test.out", test_text, "utf-8-sig")

        # File must have been written
        files = [f for f in os.listdir(".") if os.path.isfile(f)]
        assert "test.out" in files

        # Encoding must match the one we've given
        encoding = fs.detect_encoding("test.out")
        assert encoding == "utf-8-sig"

        # File contents must work
        contents = fs.read_file("test.out")
        assert test_text == contents


def test_valid_pathname():
    # A URL should not be valid
    from prance.util.fs import is_pathname_valid

    assert False == is_pathname_valid("\x00o.bar.org")

    # However, the current path should be.
    import os

    assert True == is_pathname_valid(os.getcwd())

    # Can't put non-strings into this function
    assert True == is_pathname_valid("foo")
    assert False == is_pathname_valid(123)

    # Can't accept too long components
    assert False == is_pathname_valid("a" * 256)
